@@ -8,32 +8,32 @@ import pathlib
 import select
 import socket
 import struct
-import subprocess
 import sys
 import tty
 import termios
 from datetime import datetime
-from io import IOBase
-from typing import List, Set, Dict, Optional, Tuple
+from typing import List, Set, Dict, Optional, Tuple, IO
 import pkg_resources
 import docker
 import docker.errors
-from .dockerapi_fixes import CustomContainerApiMixin
-
-from cincan import registry
-from cincan.command_inspector import CommandInspector
-from cincan.command_log import CommandLog, FileLog, CommandLogWriter, CommandLogIndex, CommandRunner, quote_args
+from requests.exceptions import ConnectionError
+from urllib.parse import urlparse
+from cincanregistry import list_handler, create_list_argparse, ToolRegistry, Remotes
+from cincanregistry.utils import parse_file_time, format_time
+from cincan.command_log import CommandLog, FileLog, CommandLogWriter, CommandRunner, quote_args
 from cincan.configuration import Configuration
 from cincan.container_check import ContainerCheck
 from cincan.file_tool import FileResolver, FileMatcher
 from cincan.tar_tool import TarTool
-
-DEFAULT_IMAGE_FILTER_TAG = "latest-stable"
+from cincan.image_fetcher import ImageFetcher
+from docker.utils import kwargs_from_env
+from cincan.version_handler import VersionHandler
 
 
 class ToolStream:
     """Handle stream to or from the container"""
-    def __init__(self, stream: IOBase):
+
+    def __init__(self, stream: IO):
         self.data_length = 0
         self.hash = hashlib.sha256()
         self.raw = bytearray()  # when collected
@@ -51,13 +51,26 @@ class ToolImage(CommandRunner):
                  image: Optional[str] = None,
                  pull: bool = False,
                  tag: Optional[str] = None,
-                 rm: bool = True):
+                 rm: bool = True,
+                 batch: bool = False):
         self.config = Configuration()
+        self.registry = ToolRegistry()
+        # Init logger, check naming convention of "name" and "image"
         self.logger = logging.getLogger(name)
-        # FIXME dirty hack to override get_archive method. Get fixed in upstream??
-        docker.api.container.ContainerApiMixin.get_archive = CustomContainerApiMixin.get_archive
+        name, image = self.namespace_conversion(name, image)
         self.client = docker.from_env()
+        try:
+            # Attempt to configure automatically
+            kwargs = kwargs_from_env()
+            self.low_level_client = docker.APIClient(version="auto", **kwargs)
+        except:
+            self.logger.warning(
+                "Unable to configure low-level API automatically. Some properties disabled.")
+            self.low_level_client = None
+        if not self._is_docker_running():
+            sys.exit(1)
         self.loaded_image = False  # did we load the image?
+        self.batch = batch  # Use batch to disable some properties when running inside script or other automation
         if path is not None:
             self.name = name or path
             if tag is not None:
@@ -69,28 +82,26 @@ class ToolImage(CommandRunner):
         elif image is not None:
             self.name = name or image
             self.loaded_image = True
-            if pull:
-                # pull first
-                self.logger.info(f"pulling image...")
-                self.__get_image(image, pull=True)
-            else:
-                # just get the image
-                try:
-                    self.__get_image(image, pull=False)
-                except docker.errors.ImageNotFound:
-                    # image not found, try to pull it
-                    self.logger.info(f"pulling image...")
-                    self.__get_image(image, pull=True)
+            fetcher = ImageFetcher(self.config, self.registry, self.client, self.low_level_client, self.logger, self.batch)
+            self.image = fetcher.get_image(image, pull)
             self.context = '.'  # not really correct, but will do
         else:
             sys.exit("No file nor image specified")
+        self.version_handler = VersionHandler(self.config, self.registry, self.image,
+                                              self.name.rsplit(":", 1)[0], self.logger)
+        if self.config.show_updates:
+            # Only check versions if not defined to run inside script or logging level is low
+            if not self.batch and self.logger.getEffectiveLevel() < logging.WARNING:
+                self.version_handler.compare_versions()
         self.input_tar: Optional[str] = None  # use '-' for stdin
         self.input_filters: Optional[List[FileMatcher]] = None
         self.output_tar: Optional[str] = None  # use '-' for stdout
         self.output_dirs: List[str] = []  # output directories to create and download (filled with troves of data)
+        self.implicit_output: bool = True  # implicitly detect output files from working directory?
+        self.explicit_output: List[str] = []  # explicitly give the download files/dirs
         self.upload_stats: Dict[str, List] = {}  # upload file stats
         self.output_filters: Optional[List[FileMatcher]] = None
-        self.no_defaults: bool = False # If set true, ignoring container specific rules from .cincanignore
+        self.no_defaults: bool = False  # If set true, ignoring container specific rules from .cincanignore
 
         self.network_mode: Optional[str] = None  # docker run --network=<value>
         self.user: Optional[str] = None  # docker run --user=<value>
@@ -107,6 +118,45 @@ class ToolImage(CommandRunner):
         self.download_files: List[str] = []
         self.buffer_output = False
 
+    def namespace_conversion(self, name: str, image: str) -> Tuple[str, str]:
+        """
+        Method for migrating images from Docker Hub into default (Quay Container Registry at the moment)
+        Converts Docker Hub namespace into Quay Namespace and notifies user.
+        Change the name of the logger.
+        Needed for consistent version information and to avoid Docker Hub rate limits
+        """
+        if self.registry.default_remote == Remotes.DOCKERHUB:
+            # Default prefix for dockerhub: cincan
+            # DockerHub set as default - no need for conversion
+            if not image.startswith(f"{self.registry.remote_registry.full_prefix}/"):
+                self.logger.debug("Not cincan tool - do nothing.")
+            else:
+                self.logger.warning(f"Using Docker Hub for image and version source. Rate limits will be applied.")
+        else:
+            # Default is other than Docker Hub
+            if not image.startswith(f"{self.registry.remote_registry.full_prefix}/"):
+                if image.startswith('cincan/'):
+                    tool_basename = os.path.basename(image)
+                    # Convert Docker Hub cincan image to point to default registry
+                    image = f"{self.registry.remote_registry.full_prefix}/{tool_basename}"
+                    if name and name.startswith('cincan/'):
+                        name = f"{self.registry.remote_registry.full_prefix}/{tool_basename}"
+                        self.logger = logging.getLogger(name)
+                    self.logger.debug(f"We are migrating away from Docker Hub - using "
+                                      f"{self.registry.remote_registry.registry_name} as default.")
+            else:
+                self.logger.debug("Not cincan tool - do nothing.")
+        return name, image
+
+    def _is_docker_running(self):
+        """Check if Docker is working properly"""
+        try:
+            self.client.ping()
+            return True
+        except ConnectionError:
+            self.logger.error("Failed to connect to Docker Server. Is it running and with proper permissions?")
+            return False
+
     def get_tags(self) -> List[str]:
         """List image tags"""
         return self.image.tags
@@ -116,14 +166,7 @@ class ToolImage(CommandRunner):
 
     def get_creation_time(self) -> datetime:
         """Get image creation time"""
-        return registry.parse_json_time(self.image.attrs['Created'])
-
-    def __get_image(self, image: str, pull: bool = False):
-        """Get Docker image, possibly pulling it first"""
-        if pull:
-            name_tag = image.rsplit(':', 1) if ':' in image else [image, 'latest']
-            self.client.images.pull(name_tag[0], tag=name_tag[1])
-        self.image = self.client.images.get(image)
+        return parse_file_time(self.image.attrs['Created'])
 
     def __create_container(self, upload_files: Dict[pathlib.Path, str], input_files: List[FileLog]):
         """Create a container from the image here"""
@@ -319,7 +362,16 @@ class ToolImage(CommandRunner):
             if log.exit_code == 0:
                 # download results
                 tar_tool = TarTool(self.logger, container, self.upload_stats, explicit_file=self.output_tar)
-                log.out_files.extend(tar_tool.download_files(self.output_filters, self.no_defaults))
+                if self.explicit_output:
+                    # just use the explicitly given output
+                    dn_files = tar_tool.download_files(self.output_filters, self.no_defaults,
+                                                       file_paths=self.explicit_output, implicit_output=False)
+                else:
+                    # try to implicitly resolve files
+                    dn_files = tar_tool.download_files(self.output_filters, self.no_defaults,
+                                                       file_paths=self.output_dirs,
+                                                       implicit_output=self.implicit_output)
+                log.out_files.extend(dn_files)
         finally:
             self.logger.debug("killing the container")
             container.kill()
@@ -365,7 +417,6 @@ def image_default_args(sub_parser):
     sub_parser.add_argument('tool', help="the tool and possible arguments", nargs=argparse.REMAINDER)
     sub_parser.add_argument('-p', '--path', help='path to Docker context')
     sub_parser.add_argument('-u', '--pull', action='store_true', help='Pull image from registry')
-
     sub_parser.add_argument('--in', dest='input_tar', nargs='?',
                             help='Provide the input files to load unfiltered into the container working directory')
     sub_parser.add_argument('--out', dest='output_tar', nargs='?',
@@ -377,7 +428,12 @@ def image_default_args(sub_parser):
     sub_parser.add_argument('-d', '--mkdir', action='append', dest='output_dir', nargs='?',
                             help='Force an empty directory to container')
     sub_parser.add_argument('--no-defaults', action='store_true',
-                            help='Ignore all container specific output filters. (Defined inside container in .cincanignore file)')
+                            help='Ignore all container specific output filters. '
+                                 '(Defined inside container in .cincanignore file)')
+    sub_parser.add_argument('-M', '--no-implicit-output', action='store_true',
+                            help='No implicit output file detection, use the directories by --mkdir')
+    sub_parser.add_argument('-e', '--explicit-output', action='append', dest='explicit_output', nargs='?',
+                            help='Specify output files/directories to download explicitly')
 
     # Docker look-a-like settings for 'cincan run'
 
@@ -395,12 +451,6 @@ def image_default_args(sub_parser):
     sub_parser.add_argument('-t', '--tty', action='store_true', help='Allocate a pseudo-TTY (see docker run --help)')
 
 
-def docker_connect_error():
-    print("Couldn't connect to Docker, is it running & user has perms to access socket?",
-          file=sys.stderr)
-    sys.exit(1)
-
-
 def get_version_information():
     """Return version of currently installed 'cincan-command' tool."""
     pkg_name = "cincan-command"
@@ -416,9 +466,21 @@ def get_version_information():
 
 def main():
     """Parse command line and run the tool"""
-    m_parser = argparse.ArgumentParser()
+    description_text = '''\
+  CinCan Command - https://gitlab.com/CinCan/cincan-command/\n
+  For full documentation, see: https://cincan.gitlab.io/cincan-command/
+    '''
+    epilog = '''  Report issues at https://gitlab.com/CinCan/cincan-command/-/issues
+    '''
+    m_parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter,
+                                       description=description_text, epilog=epilog)
+
     m_parser.add_argument("-l", "--log", dest="log_level", choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'],
                           help="Set the logging level", default=None)
+    m_parser.add_argument('--batch', action='store_true', help='Use with automation. Disables some '
+                                                               'properties meant for interactive tty device(s): '
+                                                               'Version checking disabled, '
+                                                               'pull-progress-bar disabled.')
     m_parser.add_argument('-q', '--quiet', action='store_true', help='Be quite quiet')
     m_parser.add_argument('-v', '--version', action='store_true', help='Shows currently installed version of the tool.')
     subparsers = m_parser.add_subparsers(dest='sub_command')
@@ -429,27 +491,11 @@ def main():
     test_parser = subparsers.add_parser('test')
     image_default_args(test_parser)
 
-    list_parser = subparsers.add_parser('list', formatter_class=argparse.ArgumentDefaultsHelpFormatter)
-    list_exclusive_group = list_parser.add_mutually_exclusive_group()
-    list_exclusive_group.add_argument('-t', '--tag', default=DEFAULT_IMAGE_FILTER_TAG, help='Filter images by tag name.')
-    list_exclusive_group.add_argument('-a', '--all', action='store_true', help='List all images from the registry.')
-    list_parser.add_argument('-w', '--with-tags', action='store_true', help='Show all tags of selected images.')
+    list_parser = create_list_argparse(subparsers)
 
     mani_parser = subparsers.add_parser('manifest')
     image_default_args(mani_parser)
-
-    fanin_parser = subparsers.add_parser('fanin', help='Show fan-in to the given file')
-    fanin_parser.add_argument('file', help="File to analyze")
-    fanin_parser.add_argument('-d', '--max-depth', default=3, help='Maximum tree depth')
-    fanout_parser = subparsers.add_parser('fanout', help='Show fan-out from the given file')
-    fanout_parser.add_argument('file', help="File to analyze")
-    fanout_parser.add_argument('-d', '--max-depth', default=3, help='Maximum tree depth')
-
     help_parser = subparsers.add_parser('help')
-
-    commit_parser = subparsers.add_parser('commit')
-    image_default_args(commit_parser)
-
     if len(sys.argv) > 1:
         args = m_parser.parse_args(args=sys.argv[1:])
     else:
@@ -467,19 +513,26 @@ def main():
         m_parser.print_help()
         sys.exit(1)
     elif sub_command in {'run', 'test'}:
+        # We do not want informative version logs here unless DEBUG mode
+        if logging.DEBUG < logging.getLogger().getEffectiveLevel() < logging.ERROR:
+            logging.getLogger('versions').setLevel(logging.ERROR)
+            # Also suppress meta handler output
+            logging.getLogger('metahandler').setLevel(logging.ERROR)
         if len(args.tool) == 0:
             sys.exit('Missing tool name argument')
         name = args.tool[0]
         if args.path is None:
-            tool = ToolImage(name, image=name, pull=args.pull)
+            tool = ToolImage(name, image=name, pull=args.pull, batch=args.batch)
         elif args.path is not None:
-            tool = ToolImage(name, path=args.path)
+            tool = ToolImage(name, path=args.path, batch=args.batch)
         else:
-            tool = ToolImage(name)  # should raise exception
+            tool = ToolImage(name, batch=args.batch)  # should raise exception
 
         tool.input_tar = args.input_tar if args.input_tar else None
         tool.output_tar = args.output_tar if args.output_tar else None
         tool.output_dirs = args.output_dir or []
+        tool.implicit_output = not args.no_implicit_output
+        tool.explicit_output = args.explicit_output
         tool.input_filters = FileMatcher.parse(args.in_filter) if args.in_filter is not None else None
         tool.output_filters = FileMatcher.parse(args.out_filter) if args.out_filter is not None else None
         tool.no_defaults = args.no_defaults if args.no_defaults else False
@@ -498,7 +551,7 @@ def main():
         all_args = args.tool[1:]
         if sub_command == 'test':
             check = ContainerCheck(tool)
-            tool.logger.info("# {} {}".format(','.join(tool.get_tags()), registry.format_time(tool.get_creation_time())))
+            tool.logger.info("# {} {}".format(','.join(tool.get_tags()), format_time(tool.get_creation_time())))
             log = check.run(all_args)
         else:
             log = tool.run(all_args)
@@ -512,63 +565,17 @@ def main():
         if log.stderr:
             sys.stderr.buffer.write(log.stderr)
         sys.exit(log.exit_code)  # exit code
-    elif sub_command in {'fanin', 'fanout'}:
-        inspector = CommandInspector(CommandLogIndex(), pathlib.Path().resolve())
-        depth = int(args.max_depth)
-        if sub_command == 'fanout':
-            res = inspector.fanout(pathlib.Path(args.file).resolve(), depth)
-        else:
-            res = inspector.fanin(pathlib.Path(args.file).resolve(), depth)
-        print(res)
     elif sub_command == 'manifest':
         # sub command 'manifest'
         if len(args.tool) == 0:
             sys.exit('Missing tool name argument')
         name = args.tool[0]
-        reg = registry.ToolRegistry()
-        try:
-            info = reg.fetch_manifest(name)
-        except OSError:
-            docker_connect_error()
-        print(json.dumps(info, indent=2))
+        reg = ToolRegistry()
+        conf = Configuration()
+        name, tag = name.rsplit(":", 1) if ":" in name else [name, conf.default_stable_tag]
+        info = reg.remote_registry.fetch_manifest(name, tag)
+        print(info)
     elif sub_command == 'list':
-        format_str = "{0:<25}"
-        if args.with_tags:
-            format_str += " {4:<30}"
-        format_str += " {1}"
-        reg = registry.ToolRegistry()
-        try:
-            tool_list = reg.list_tools(defined_tag=args.tag if not args.all else None)
-        except OSError:
-            docker_connect_error()
-        if not args.all and tool_list:
-            print(f"\n  Listing all tools with tag '{args.tag}':\n")
-        for tool in sorted(tool_list):
-            lst = tool_list[tool]
-            print(format_str.format(lst.name, lst.description, ",".join(lst.input), ",".join(lst.output),
-                                    ",".join(lst.tags)))
-    elif sub_command == 'commit':
-
-        log_path = str(pathlib.Path.home() / '.cincan/shared')
-
-        # change working dir where logs lie.
-        os.chdir(log_path)
-
-        print("check if git exists in current directory")
-
-        if os.path.exists('.git'):
-            print("if git exists, pull repo")
-            subprocess.call(["git", "pull"])
-            print("Add, commit and push logs")
-            subprocess.call(["git", "add", "."])
-            subprocess.call(["git", "status"])
-            subprocess.call(["git", "commit", "-m", "added log files from shared folder with cincan commit -command"])
-            subprocess.call(["git", "push"])      
-        else:
-            print("Git doesn't exist. If you want to share your logs: Go to .cincan/shared folder")
-            print("type 'git init' and attach folder to remote repository for sharing logs")
-            print("git remote add origin git@gitlab.com:CinCan/log-sharing.git")
-            print("git pull origin master")
-            print("git branch --set-upstream-to=origin/master master")
+        list_handler(args)
     else:
         sys.exit(f"Unexpected sub command '{sub_command}")
